//##################################################################
// DumpModules (builds in either x86 or x64 mode)
//##################################################################

#include DumpModules.h

DATAFORCALLBACK STRUCT
	pMapFile PTR
	hListview HANDLE
	Pass DD
ends

DATA SECTION
	hInstance	HANDLE		?

CODE SECTION

START:
	invoke GetModuleHandle, 0
	mov [hInstance],rax
	invoke InitCommonControls
	INVOKE DialogBoxParam,[hInstance],1000,0,OFFSET DlgProc,0
	invoke ExitProcess,0

DlgProc FRAME hwnd,uMsg,wParam,lParam
	uses rbx
	LOCAL rc:RECT

	cmp D[uMsg],WM_INITDIALOG
	jne >>.WM_COMMAND
		invoke GetClientRect,[hwnd],offset rc
		invoke GetDlgItem,[hwnd],1001
		mov rbx,rax

		invoke SetWindowPos,rbx,NULL,[rc.left],[rc.top],[rc.right],[rc.bottom],SWP_NOZORDER

		invoke SetupDllListView,rbx
		invoke WalkLoadedModules,rbx

		// Sort the listview based on the import type
		invoke SendMessage,rbx,LVM_SORTITEMSEX,rbx,OFFSET ListViewSortProc

		jmp >.EXIT

	.WM_COMMAND
	cmp D[uMsg],WM_COMMAND
	jne >.WM_CLOSE

		jmp >.EXIT

	.WM_CLOSE
	cmp D[uMsg],WM_CLOSE
	jne >.DEFPROC
		INVOKE EndDialog,[hwnd],0

	.DEFPROC
		mov rax,FALSE
		ret

	.EXIT

	mov rax, TRUE
	ret
ENDF

MapPEFile FRAME pszFilename, phFile, phMapfile, ppMapfile
	uses rbx,rdi,rsi

	// Get a file handle
	invoke CreateFile,[pszFilename],GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,NULL,NULL
	mov rbx,[phFile]
	mov [rbx],rax
	test rax,rax
	jns >
		ret
	:

	// Map the file
	invoke CreateFileMapping,[rbx],NULL,PAGE_READONLY,0,0,NULL
	mov rdi,[phMapfile]
	mov [rdi],rax

	// Check the return value of CreateFileMapping to see if there was an error
	test rax,rax
	jnz >
		invoke CloseHandle,[rbx]
		xor rax,rax
		dec rax
		ret
	:

	// Get the address of the mapping
	invoke MapViewOfFile,[rdi],FILE_MAP_READ,0,0,0
	mov rsi,[ppMapfile]
	mov [rsi],rax

	// Check the return value of MapViewOfFile to see if there was an error
	test rax,rax
	jnz >
		invoke CloseHandle,[rdi]
		invoke CloseHandle,[rbx]
		xor rax,rax
		dec rax
		ret
	:

	xor rax,rax
	ret

ENDF

UmmapPEFile FRAME hFile, hMapfile, pMapfile
	// Unmap the file and close the handles
	invoke UnmapViewOfFile,[pMapfile]
	invoke CloseHandle,[hMapfile]
	invoke CloseHandle,[hFile]
	RET
ENDF

WalkLoadedModules FRAME hListview
	uses rbx,rdi,rsi
	LOCAL hProcess:%HANDLE
	LOCAL cbNeeded:D
	LOCAL BaseName[MAX_PATH]:%CHAR
	LOCAL hFile:%HANDLE
	LOCAL hMapfile:%HANDLE
	LOCAL pMapfile:%PTR
	LOCAL dfc:DATAFORCALLBACK

	// Get this processes executable name and map the PE file
	invoke GetModuleFileName,NULL,offset BaseName,MAX_PATH
	invoke MapPEFile,offset BaseName,offset hFile,offset hMapfile,offset pMapfile
	// Be sure the file was mapped
	test eax,eax
	jns >
		// Skip everything if we can't map the file
		ret
	:

	// This is always -1 but just in case Microsoft ever changes it
	invoke GetCurrentProcess
	mov [hProcess],rax

	// Since more than one thing has to be passed to the
	// callback we'll use a structure and pass its address
	// this gives us more flexibility.
	mov rax,[pMapfile]
	mov [dfc.pMapFile],rax
	mov rax,[hListview]
	mov [dfc.hListview],rax
	mov D[dfc.Pass],0

	invoke EnumerateLoadedModules64,[hProcess],offset EnumerateLoadedModulesProc,offset dfc

	invoke UmmapPEFile,[hFile],[hMapfile],[pMapfile]
	ret
endf

// Because the ModuleBase parameter in the callback is 64 bit in both X64 and X86 builds
// we need a different set of parameters for each build type
#IF X64
EnumerateLoadedModulesProc FRAME ModuleName, ModuleBase, ModuleSize, UserContext
#ELSE
EnumerateLoadedModulesProc FRAME ModuleName, ModuleBase, ModuleBaseHigh, ModuleSize, UserContext
#ENDIF
	uses rsi,rdi,rbx

	mov rdi,[UserContext]

	cmp D[rdi+DATAFORCALLBACK.Pass],0
	jne >
		// skip the first enumeration, its this processes exe file
		inc D[rdi+DATAFORCALLBACK.Pass]
		ret
	:
	inc D[rdi+DATAFORCALLBACK.Pass]

	mov rsi,offset szExternal
	invoke PathFindFileName,[ModuleName]

	mov rbx,rax
	invoke FindModuleInPE,[edi+DATAFORCALLBACK.pMapFile],rbx
	test rax,rax
	js >.EXTRN
		mov rsi,offset szInternal
	.EXTRN

	mov B[rbx-1],0

	// Add this module to the listview
	invoke AddToListview,[rdi+DATAFORCALLBACK.hListview],rbx,[ModuleBase],rsi,[ModuleName]
	mov rax,TRUE
	ret

	szExternal: CHAR A"External",0
	szInternal: CHAR A"Direct",0
endf

FindModuleInPE FRAME pMapFile, pFindFile
	uses rdi,rsi,rbx
	LOCAL cbSize:D

	// Get the address of the IMAGE_NT_HEADERS structure
	invoke ImageNtHeader,[pMapFile]
	mov rsi,rax

	invoke ImageDirectoryEntryToData,[pMapFile],FALSE,IMAGE_DIRECTORY_ENTRY_IMPORT,offset cbSize
	mov rdi,rax
	mov rbx,[cbSize]

	L1:
		// Walk the PE import directory
		invoke ImageRvaToVa,rsi,[pMapFile],[rdi+IMAGE_IMPORT_DESCRIPTOR.Name],NULL
		// Is this the module we're looking for
		invoke lstrcmpi,[pFindFile],rax
		test rax,rax
		jnz >
		// The module was found, return 0
		xor rax,rax
		ret
		:
		// Next descriptor
		add rdi,sizeof IMAGE_IMPORT_DESCRIPTOR
		sub rbx,sizeof IMAGE_IMPORT_DESCRIPTOR
		jnz <<L1

	xor rax,rax
	dec rax
	ret
endf

AddToListview FRAME hListview,pszModule,LoadAddr,pszType,pszFolder
	LOCAL lvi:LVITEM
	LOCAL szLoadAddr[64]:%CHAR

	// Standard routine to add an item to the listview

	mov D[lvi.mask],LVIF_TEXT
	mov D[lvi.iItem],0
	mov D[lvi.iSubItem],0
	mov rax,[pszModule]
	mov [lvi.pszText],rax
	invoke SendMessage,[hListview],LVM_INSERTITEM,0,offset lvi

	mov D[lvi.mask],LVIF_TEXT
	mov D[lvi.iItem],0
	mov D[lvi.iSubItem],1
	invoke wsprintf,offset szLoadAddr,"%p",[LoadAddr]
	#IF !X64
		// For 32 bit builds we have to reset ESP
		add esp,12
	#ENDIF
	lea rax,szLoadAddr
	mov [lvi.pszText],rax
	invoke SendMessage,[hListview],LVM_SETITEM,0,offset lvi

	mov D[lvi.mask],LVIF_TEXT
	mov D[lvi.iItem],0
	mov D[lvi.iSubItem],2
	mov rax,[pszType]
	mov [lvi.pszText],rax
	invoke SendMessage,[hListview],LVM_SETITEM,0,offset lvi

	mov D[lvi.mask],LVIF_TEXT
	mov D[lvi.iItem],0
	mov D[lvi.iSubItem],3
	mov rax,[pszFolder]
	mov [lvi.pszText],rax
	invoke SendMessage,[hListview],LVM_SETITEM,0,offset lvi

	ret
endf

SetupDllListView FRAME hListview
	uses edi,esi
	LOCAL LVColumn:LVCOLUMN
	LOCAL hFont:%HANDLE

	// Use a fixed font, makes the addresses look nicer
	invoke GetStockObject,ANSI_FIXED_FONT
	mov [hFont],rax

	invoke SendMessage,[hListview],WM_SETFONT,[hFont],TRUE

	mov D[LVColumn.mask],LVCF_FMT + LVCF_TEXT + LVCF_WIDTH
	mov D[LVColumn.fmt],LVCFMT_LEFT
	mov D[LVColumn.cx],200
	push ADDR "Module name"
	pop [LVColumn.pszText]
	mov D[LVColumn.cchTextMax],NULL
	invoke SendMessage,[hListview],LVM_INSERTCOLUMN,0,OFFSET LVColumn

	mov D[LVColumn.mask],LVCF_FMT + LVCF_TEXT + LVCF_WIDTH
	mov D[LVColumn.fmt],LVCFMT_LEFT
	mov D[LVColumn.cx],150
	push ADDR "Load Address"
	pop [LVColumn.pszText]
	mov D[LVColumn.cchTextMax],NULL
	invoke SendMessage,[hListview],LVM_INSERTCOLUMN,1,OFFSET LVColumn

	mov D[LVColumn.mask],LVCF_FMT + LVCF_TEXT + LVCF_WIDTH
	mov D[LVColumn.fmt],LVCFMT_LEFT
	mov D[LVColumn.cx],80
	push ADDR "Type"
	pop [LVColumn.pszText]
	mov D[LVColumn.cchTextMax],NULL
	invoke SendMessage,[hListview],LVM_INSERTCOLUMN,2,OFFSET LVColumn

	mov D[LVColumn.mask],LVCF_FMT + LVCF_TEXT + LVCF_WIDTH
	mov D[LVColumn.fmt],LVCFMT_LEFT
	mov D[LVColumn.cx],150
	push ADDR "Folder"
	pop [LVColumn.pszText]
	mov D[LVColumn.cchTextMax],NULL
	invoke SendMessage,[hListview],LVM_INSERTCOLUMN,3,OFFSET LVColumn

	invoke SendMessage,[hListview],LVM_SETEXTENDEDLISTVIEWSTYLE,LVS_EX_GRIDLINES + LVS_EX_FULLROWSELECT,LVS_EX_FULLROWSELECT + LVS_EX_GRIDLINES

	ret
ENDF

ListViewSortProc FRAME lParam1,lParam2,lParamSort
	uses ebx,edi,esi
	LOCAL	ItemData1		:LVITEM
	LOCAL	ItemData2		:LVITEM
	LOCAL	Tbuff1[256]		:%CHAR
	LOCAL	Tbuff2[256]		:%CHAR

	mov rax,[lParam1]
	mov [ItemData1.iItem],rax
	mov D[ItemData1.iSubItem],2
	mov D[ItemData1.mask],LVIF_TEXT
	lea rax,Tbuff1
	mov [ItemData1.pszText],rax
	mov D[ItemData1.cchTextMax],255
	invoke SendMessage,[lParamSort],LVM_GETITEM,0,OFFSET ItemData1

	mov rax,[lParam2]
	mov [ItemData2.iItem],rax
	mov D[ItemData2.iSubItem],2
	mov D[ItemData2.mask],LVIF_TEXT
	lea rax,Tbuff2
	mov [ItemData2.pszText],rax
	mov D[ItemData2.cchTextMax],255
	invoke SendMessage,[lParamSort],LVM_GETITEM,0,OFFSET ItemData2

	invoke lstrcmpi,offset Tbuff1,offset Tbuff2
	RET
ENDF
